"""
AyxPlugin (required) has-a IncomingInterface (optional).
Although defining IncomingInterface is optional, the interface methods are needed if an upstream tool exists.
"""

import AlteryxPythonSDK as Sdk
import engine.sf_authorization as sf_authorization
import engine.sf_datapush as sf_datapush
import engine.sf_metadata as sf_metadata
import engine.sf_push_metadata as sf_push_metadata
import engine.sf_queries as sf_queries
import engine.sf_reports as sf_reports
import engine.sf_utils as sf_utils
import pickle
import json
import aiohttp
import xml.etree.ElementTree as Et

class AyxPlugin:
    """
    Implements the plugin interface methods, to be utilized by the Alteryx engine to communicate with this plugin.
    Prefixed with "pi", the Alteryx engine will expect the below five interface methods to be defined.
    """

    def __init__(self, n_tool_id: int, alteryx_engine: object, output_anchor_mgr: object):
        """
        Constructor is called whenever the Alteryx engine wants to instantiate an instance of this plugin.
        :param n_tool_id: The assigned unique identification for a tool instance.
        :param alteryx_engine: Provides an interface into the Alteryx engine.
        :param output_anchor_mgr: A helper that wraps the outgoing connections for a plugin.
        """

        # Default properties
        self.n_tool_id = n_tool_id
        self.alteryx_engine = alteryx_engine
        self.output_anchor_mgr = output_anchor_mgr

        # Custom properties
        self.api_version = 'v42.0' 
        self.output_anchor = None
        self.is_initialized = True
        self.update_only = self.alteryx_engine.get_init_var(self.n_tool_id, 'UpdateOnly') == 'False'
        self.header_list = None
        self.query_type = None

    def pi_init(self, str_xml: str):
        """
        Handles extracting user-entered file name, and input data verification.
        Called when the Alteryx engine is ready to provide the tool configuration from the GUI.
        :param str_xml: The raw XML from the GUI.
        """
        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')

        self.tool_guid = sf_utils.get_xml_config_value(str_xml, 'tool_guid')
        self.temp_file_name = str(self.tool_guid) + str(self.n_tool_id) #metadata file name 
        self.oauth_endpoint = sf_utils.get_xml_config_value(str_xml, 'oauthEndpoint') if 'oauthEndpoint' in str_xml else 'https://login.salesforce.com/services/oauth2/token'

        self.username = sf_utils.get_xml_config_value(str_xml, 'username')
        self.password = sf_utils.get_xml_config_value(str_xml, 'password')
        # Error checking for missing username or password
        if self.username == None:
            self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.error, self.alteryx_engine.xmsg("Enter a username"))
            return
        if self.password == None:
            self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.error, self.alteryx_engine.xmsg("Enter a password"))
            return
        
        self.password = self.alteryx_engine.decrypt_password(self.password) # use with HTML GUI SDK v2

        self.security_token = sf_utils.get_xml_config_value(str_xml, 'securityToken')

        if self.security_token != None:
            self.security_token = self.alteryx_engine.decrypt_password(self.security_token)

        self.instance_url = sf_utils.get_xml_config_value(str_xml, 'instance_url')
        self.access_token = sf_utils.get_xml_config_value(str_xml, 'access_token')
        self.query = sf_utils.get_xml_config_value(str_xml, 'query')
        self.json_parse = sf_utils.get_xml_config_value(str_xml, 'ParseJSON')
        self.report_parse = sf_utils.get_xml_config_value(str_xml, 'ParseReport')
        self.query_type = sf_utils.get_xml_config_value(str_xml, 'queryType')
        self.report_id = sf_utils.get_xml_config_value(str_xml, 'report')
        self.page = sf_utils.get_xml_config_value(str_xml, 'page')
        self.table = sf_utils.get_xml_config_value(str_xml, 'table')
        self.output_fields = sf_utils.get_xml_config_value(str_xml, 'outputFields')
        metadata_string = sf_utils.get_xml_config_value(str_xml, 'reduced_metadata')
        if metadata_string:
            self.metadata_list = json.loads(metadata_string)

        # Error checking for missing table and output fields when using Query Builder
        if self.page == 'queryTypeSelector' and self.query_type == 'queryBuilder' and self.table == None:
            self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.error, self.alteryx_engine.xmsg("Select a table"))
            return
        if self.page == 'queryTypeSelector' and self.query_type == 'queryBuilder' and self.output_fields == None:
            self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.error, self.alteryx_engine.xmsg("Select at least one output field"))
            return

        # Retrieves metadata
        if self.query and self.query_type != 'existingReport':
            sf_push_metadata.pass_metadata(self=self)

        if self.update_only:
            sf_authorization.get_access_token(self=self)

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> object:
        """
        The IncomingInterface objects are instantiated here, one object per incoming connection, however since
        this tool does not accept an incoming connection, instantiation is not needed and "ii" methods won't be called.
        Called when the Alteryx engine is attempting to add an incoming data connection.
        :param str_type: The name of the input connection anchor, defined in the Config.xml file.
        :param str_name: The name of the wire, defined by the workflow author.
        :return: self.
        """

        return None

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        """
        Called when the Alteryx engine is attempting to add an outgoing data connection.
        :param str_name: The name of the output connection anchor, defined in the Config.xml file.
        :return: True signifies that the connection is accepted.
        """

        return True

    def pi_push_all_records(self, n_record_limit: int) -> bool:
        """
        Handles pushing records out to downstream tool(s).
        Called when a tool has no incoming data connection.
        :param n_record_limit: Set it to <0 for no limit, 0 for no records, and >0 to specify the number of records.
        :return: False if there are issues with the input data or if the workflow isn't being ran, otherwise True.
        """

        if self.header_list and self.query_type != 'existingReport':
            self.record_info_out = sf_metadata.build_record_info_out(self=self, response_object=self.header_list)
            self.output_anchor.init(self.record_info_out)
        # Reads temp file to populate tool metadata to downstream tools when not in configuration mode for reports
        elif self.alteryx_engine.get_init_var(self.n_tool_id, 'UpdateMode') == 'Quick' and self.query_type == 'existingReport':
            try:
                file_path =  self.alteryx_engine.create_temp_file_name(extension=str(self.temp_file_name), options=2)
                with open(file_path, 'rb') as fp:
                    header_list = pickle.load(fp)
                record_info_out = sf_metadata.build_record_info_out(self=self, response_object=header_list)
                self.output_anchor.init(record_info_out)
            except:
                self.alteryx_engine.output_message(self.n_tool_id, Sdk.EngineMessageType.info, self.alteryx_engine.xmsg("Run workflow to produce field metadata"))

        if not self.is_initialized:
            return False

        if self.alteryx_engine.get_init_var(self.n_tool_id, 'UpdateOnly') == 'True':
            return False

        if not self.password or not self.username:
            return False
        
        if self.query_type == 'existingReport':
            sf_reports.process_report(self=self)
        else:
            try:
                sf_queries.process_query(self=self)
            except aiohttp.client_exceptions.ClientResponseError as error:
                sf_utils.display_msg(self=self, type='error', msg_string=str(str(error.status) + ': ' + error.message))
                return False

        return True

    def pi_close(self, b_has_errors: bool):
        """
        Called after all records have been processed.
        :param b_has_errors: Set to true to not do the final processing.
        """

        self.output_anchor.assert_close()  # Checks whether connections were properly closed.

class IncomingInterface:
    """
    This optional class is returned by pi_add_incoming_connection, and it implements the incoming interface methods, to
    be utilized by the Alteryx engine to communicate with a plugin when processing an incoming connection.
    Prefixed with "ii", the Alteryx engine will expect the below four interface methods to be defined.
    """

    def __init__(self, parent: object):
        """
        Constructor for IncomingInterface.
        :param parent: AyxPlugin
        """

        pass

    def ii_init(self, record_info_in: object) -> bool:
        """
        Called to report changes of the incoming connection's record metadata to the Alteryx engine.
        :param record_info_in: A RecordInfo object for the incoming connection's fields.
        """

        pass

    def ii_push_record(self, in_record: object) -> bool:
        """
        Called when an input record is being sent to the plugin.
        :param in_record: The data for the incoming record.
        """

        pass

    def ii_update_progress(self, d_percent: float):
        """
        Called by the upstream tool to report what percentage of records have been pushed.
        :param d_percent: Value between 0.0 and 1.0.
        """

        pass

    def ii_close(self):
        """
        Called when the incoming connection has finished passing all of its records.
        """

        pass
